@@ -263,7 +263,7 @@ static inline int php_pinba_parse_server(char *address, char **host, char **port
 }
 /* }}} */
 
-static inline int php_pinba_timer_stop(pinba_timer_t *t) /* {{{ */
+static inline int php_pinba_timer_stop(pinba_timer_t *t, struct timeval *pnow, struct rusage *pu) /* {{{ */
 {
 	struct timeval now;
 	struct rusage u, tmp;
@@ -272,16 +272,29 @@ static inline int php_pinba_timer_stop(pinba_timer_t *t) /* {{{ */
 		return FAILURE;
 	}
 
-	gettimeofday(&now, 0);
+	if (pnow) {
+		now = *pnow;
+	} else {
+		if (gettimeofday(&now, 0) != 0) {
+			return FAILURE;
+		}
+	}
+
 	timersub(&now, &t->start, &t->value);
 
-	if (getrusage(RUSAGE_SELF, &u) == 0) {
-		timersub(&u.ru_utime, &t->tmp_ru_utime, &tmp.ru_utime);
-		timersub(&u.ru_stime, &t->tmp_ru_stime, &tmp.ru_stime);
-		timeradd(&t->ru_utime, &tmp.ru_utime, &t->ru_utime);
-		timeradd(&t->ru_stime, &tmp.ru_stime, &t->ru_stime);
+	if (pu) {
+		u = *pu;
+	} else {
+		if (getrusage(RUSAGE_SELF, &u) != 0) {
+			return FAILURE;
+		}
 	}
 
+	timersub(&u.ru_utime, &t->tmp_ru_utime, &tmp.ru_utime);
+	timersub(&u.ru_stime, &t->tmp_ru_stime, &tmp.ru_stime);
+	timeradd(&t->ru_utime, &tmp.ru_utime, &t->ru_utime);
+	timeradd(&t->ru_stime, &tmp.ru_stime, &t->ru_stime);
+
 	t->started = 0;
 	return SUCCESS;
 }
@@ -382,7 +395,7 @@ static void php_timer_resource_dtor(zend_resource *entry) /* {{{ */
 {
 	pinba_timer_t *t = (pinba_timer_t *)entry->ptr;
 
-	php_pinba_timer_stop(t);
+	php_pinba_timer_stop(t, NULL, NULL);
 	/* php_pinba_timer_dtor(t); all timers are destroyed at once */
 
 	/* but we don't need the user data anymore */
@@ -399,17 +412,19 @@ static void php_timer_resource_dtor(zend_resource *entry) /* {{{ */
 }
 /* }}} */
 
-static int php_pinba_timer_stop_helper(zval *zv, void *arg) /* {{{ */
+static int php_pinba_timer_stop_helper(zval *zv, int num_args, va_list args, zend_hash_key *hash_key) /* {{{ */
 {
-	long flags = (long)arg;
-
 	if (Z_RES_TYPE_P(zv) == le_pinba_timer) {
+		long flags = va_arg(args, long);
+		struct timeval *now = va_arg(args, struct timeval *);
+		struct rusage *u = va_arg(args, struct rusage *);
+
 		pinba_timer_t *t = (pinba_timer_t *)Z_RES_VAL_P(zv);
 
 		if (t->deleted || ((flags & PINBA_FLUSH_ONLY_STOPPED_TIMERS) != 0 && t->started)) {
 			return ZEND_HASH_APPLY_KEEP;
 		} else {
-			php_pinba_timer_stop(t);
+			php_pinba_timer_stop(t, now, u);
 			t->deleted = 1; /* ignore next time */
 		}
 
@@ -423,6 +438,22 @@ static int php_pinba_timer_stop_helper(zval *zv, void *arg) /* {{{ */
 }
 /* }}} */
 
+static int php_pinba_timer_delete_helper(zval *zv) /* {{{ */
+{
+	if (Z_RES_TYPE_P(zv) == le_pinba_timer) {
+		pinba_timer_t *t = (pinba_timer_t *)Z_RES_VAL_P(zv);
+
+		if (t->deleted) {
+			/* only delete from the list if there are no references to this zval */
+			if (GC_REFCOUNT(Z_COUNTED_P(zv)) == 1) {
+				return ZEND_HASH_APPLY_REMOVE;
+			}
+		}
+	}
+	return ZEND_HASH_APPLY_KEEP;
+}
+/* }}} */
+
 static size_t sapi_ub_write_counter(const char *str, size_t length) /* {{{ */
 {
 	PINBA_G(tmp_req_data).doc_size += length;
@@ -730,7 +761,13 @@ static inline Pinba__Request *php_create_pinba_packet(pinba_client_t *client, co
 			old_t = zend_hash_str_find_ptr(&timers_uniq, hashed_tags, hashed_tags_len);
 			if (old_t != NULL) {
 				timeradd(&old_t->value, &t->value, &old_t->value);
-				old_t->hit_count++;
+				timeradd(&old_t->ru_utime, &t->ru_utime, &old_t->ru_utime);
+				timeradd(&old_t->ru_stime, &t->ru_stime, &old_t->ru_stime);
+				if (t->hit_count) {
+					old_t->hit_count += t->hit_count;
+				} else {
+					old_t->hit_count++;
+				}
 			} else {
 				zend_hash_str_add_ptr(&timers_uniq, hashed_tags, hashed_tags_len, t);
 			}
@@ -934,11 +971,40 @@ static inline void php_pinba_req_data_dtor(pinba_req_data *record) /* {{{ */
 }
 /* }}} */
 
+static inline void php_pinba_reset_data(void) /* {{{ */
+{
+	struct timeval t;
+	struct rusage u;
+
+	if (gettimeofday(&t, 0) == 0) {
+		timeval_cvt(&(PINBA_G(tmp_req_data).req_start), &t);
+	}
+
+	if (getrusage(RUSAGE_SELF, &u) == 0) {
+		timeval_cvt(&(PINBA_G(tmp_req_data).ru_utime), &u.ru_utime);
+		timeval_cvt(&(PINBA_G(tmp_req_data).ru_stime), &u.ru_stime);
+	}
+
+	PINBA_G(tmp_req_data).doc_size = 0;
+	PINBA_G(tmp_req_data).mem_peak_usage= 0;
+	PINBA_G(tmp_req_data).req_count = 0;
+
+	zend_hash_clean(&PINBA_G(tags));
+	zend_hash_clean(&PINBA_G(timers));
+}
+/* }}} */
+
 static void php_pinba_flush_data(const char *custom_script_name, long flags) /* {{{ */
 {
+	struct timeval now;
+	struct rusage u;
+
+	if (gettimeofday(&now, 0) != 0 || getrusage(RUSAGE_SELF, &u) != 0) {
+		return;
+	}
 
 	/* stop all running timers */
-	zend_hash_apply_with_argument(&EG(regular_list), (apply_func_arg_t) php_pinba_timer_stop_helper, (void *)flags);
+	zend_hash_apply_with_arguments(&EG(regular_list), (apply_func_args_t) php_pinba_timer_stop_helper, 3, (void *)flags, (void *)&now, (void *)&u);
 
 	/* prevent any further access to the timers */
 	PINBA_G(timers_stopped) = 1;
@@ -946,6 +1012,7 @@ static void php_pinba_flush_data(const char *custom_script_name, long flags) /*
 	if (!PINBA_G(enabled) || PINBA_G(n_collectors) == 0) {
 		/* disabled or no collectors defined, exit */
 		zend_hash_clean(&PINBA_G(timers));
+		zend_hash_apply(&EG(regular_list), (apply_func_t) php_pinba_timer_delete_helper);
 		PINBA_G(timers_stopped) = 0;
 		return;
 	}
@@ -958,25 +1025,12 @@ static void php_pinba_flush_data(const char *custom_script_name, long flags) /*
 	php_pinba_req_data_send(NULL, custom_script_name, flags);
 
 	if (flags & PINBA_FLUSH_RESET_DATA) {
-		struct timeval t;
-		struct rusage u;
-
-		if (gettimeofday(&t, 0) == 0) {
-			timeval_cvt(&(PINBA_G(tmp_req_data).req_start), &t);
-		}
-
-		if (getrusage(RUSAGE_SELF, &u) == 0) {
-			timeval_cvt(&(PINBA_G(tmp_req_data).ru_utime), &u.ru_utime);
-			timeval_cvt(&(PINBA_G(tmp_req_data).ru_stime), &u.ru_stime);
-		}
-
-		PINBA_G(tmp_req_data).doc_size = 0;
-		PINBA_G(tmp_req_data).mem_peak_usage= 0;
-		PINBA_G(tmp_req_data).req_count = 0;
-
-		zend_hash_clean(&PINBA_G(tags));
+		php_pinba_reset_data();
 	}
 
+	/* delete all stopped timers */
+	zend_hash_apply(&EG(regular_list), (apply_func_t) php_pinba_timer_delete_helper);
+
 	PINBA_G(timers_stopped) = 0;
 	zend_hash_clean(&PINBA_G(timers));
 }
@@ -1078,7 +1132,7 @@ static pinba_timer_t *php_pinba_timer_ctor(pinba_timer_tag_t **tags, int tags_nu
 }
 /* }}} */
 
-static void php_pinba_get_timer_info(pinba_timer_t *t, zval *info) /* {{{ */
+static void php_pinba_get_timer_info(pinba_timer_t *t, zval *info, struct timeval *pnow) /* {{{ */
 {
 	zval tags;
 	pinba_timer_tag_t *tag;
@@ -1088,7 +1142,12 @@ static void php_pinba_get_timer_info(pinba_timer_t *t, zval *info) /* {{{ */
 	array_init(info);
 
 	if (t->started) {
-		gettimeofday(&tmp, 0);
+		if (pnow) {
+			tmp = *pnow;
+		} else {
+			gettimeofday(&tmp, 0);
+		}
+
 		timersub(&tmp, &t->start, &tmp);
 		timeradd(&t->value, &tmp, &tmp);
 	} else {
@@ -1320,7 +1379,7 @@ static PHP_FUNCTION(pinba_timer_stop)
 		RETURN_FALSE;
 	}
 
-	php_pinba_timer_stop(t);
+	php_pinba_timer_stop(t, NULL, NULL);
 	RETURN_TRUE;
 }
 /* }}} */
@@ -1339,7 +1398,7 @@ static PHP_FUNCTION(pinba_timer_delete)
 	PHP_ZVAL_TO_TIMER(timer, t);
 
 	if (t->started) {
-		php_pinba_timer_stop(t);
+		php_pinba_timer_stop(t, NULL, NULL);
 	}
 
 	t->deleted = 1;
@@ -1539,6 +1598,14 @@ static PHP_FUNCTION(pinba_flush)
 }
 /* }}} */
 
+/* {{{ proto void pinba_reset()
+   Reset the data */
+static PHP_FUNCTION(pinba_reset)
+{
+	php_pinba_reset_data();
+}
+/* }}} */
+
 /* {{{ proto array pinba_get_info()
    Get request info */
 static PHP_FUNCTION(pinba_get_info)
@@ -1604,6 +1671,7 @@ static PHP_FUNCTION(pinba_get_info)
 	add_assoc_string(return_value, "hostname", PINBA_G(host_name));
 
 	array_init(&timers);
+	gettimeofday(&tmp, 0);
 
 	for (zend_hash_internal_pointer_reset_ex(&EG(regular_list), &pos);
 			(zv = zend_hash_get_current_data_ex((&EG(regular_list)), &pos)) != NULL;
@@ -1616,7 +1684,7 @@ static PHP_FUNCTION(pinba_get_info)
 				continue;
 			}
 
-			php_pinba_get_timer_info(t, &timer_info);
+			php_pinba_get_timer_info(t, &timer_info, &tmp);
 			add_next_index_zval(&timers, &timer_info);
 		}
 	}
@@ -1679,7 +1747,7 @@ static PHP_FUNCTION(pinba_timer_get_info)
 
 	PHP_ZVAL_TO_TIMER(timer, t);
 
-	php_pinba_get_timer_info(t, return_value);
+	php_pinba_get_timer_info(t, return_value, NULL);
 }
 /* }}} */
 
@@ -1690,11 +1758,17 @@ static PHP_FUNCTION(pinba_timers_stop)
 	zval *zv;
 	pinba_timer_t *t;
 	HashPosition pos;
+	struct timeval now;
+	struct rusage u;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "") != SUCCESS) {
 		return;
 	}
 
+	if (gettimeofday(&now, 0) != 0 || getrusage(RUSAGE_SELF, &u) != 0) {
+		RETURN_FALSE;
+	}
+
 	for (zend_hash_internal_pointer_reset_ex(&EG(regular_list), &pos);
 			(zv = zend_hash_get_current_data_ex((&EG(regular_list)), &pos)) != NULL;
 			zend_hash_move_forward_ex(&EG(regular_list), &pos)) {
@@ -1702,7 +1776,7 @@ static PHP_FUNCTION(pinba_timers_stop)
 
 		if (rsrc->type == le_pinba_timer) {
 			t = (pinba_timer_t *)rsrc->ptr;
-			php_pinba_timer_stop(t);
+			php_pinba_timer_stop(t, &now, &u);
 		}
 	}
 	RETURN_TRUE;
@@ -2297,6 +2371,9 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_pinba_flush, 0, 0, 0)
 	ZEND_ARG_INFO(0, flags)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arginfo_pinba_reset, 0, 0, 0)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(arginfo_pinba_get_info, 0, 0, 0)
 ZEND_END_ARG_INFO()
 
@@ -2366,6 +2443,7 @@ zend_function_entry pinba_functions[] = {
 	PINBA_FUNC(pinba_timer_tags_merge)
 	PINBA_FUNC(pinba_timer_tags_replace)
 	PINBA_FUNC(pinba_flush)
+	PINBA_FUNC(pinba_reset)
 	PINBA_FUNC(pinba_get_info)
 	PINBA_FUNC(pinba_get_data)
 	PINBA_FUNC(pinba_timer_get_info)
